// <![CDATA[
//  GVMultiImageKnobView.m
//  GVCustomControls
//
//  Created by Vlad Gorloff on 28.05.12.
//  Copyright (c) 2012 Vlad Gorloff. All rights reserved.
//

/*
 References:
 http://katidev.com/blog/2008/03/07/how-to-create-a-custom-control-with-nsview/
 https://developer.apple.com/library/mac/#samplecode/AnimatedSlider/Introduction/Intro.html
 http://developer.apple.com/library/mac/#documentation/GraphicsImaging/Conceptual/CoreAnimation_Cookbook/Articles/Drawing.html 
*/

#import "GVMultiImageKnobView.h"
#import <Quartz/Quartz.h>
#import "GVCommonArithmetics.h"

static void *NormalizedValueBindingContext = (void *)@"NormalizedValue";

@interface GVMultiImageKnobView ()
{
    CGFloat mCurrentAngle;
    CGFloat mMinAngle;
    CGFloat mMaxAngle;
    CGFloat mDefaultAngle;

    // Bindings support
    id observedObjectForNormalizedValue;
    NSString *observedKeyPathForNormalizedValue;
}

@property NSColor * backgroundColor;
@property KGVMouseTrackingStatus trackingStatus;
@property NSPoint lastDragPosition;

-(void) drawView:(CGContextRef)ctx;
+ (id)defaultAnimationForKey:(NSString *)key;

-(void)postModelUpdate;

@end

@implementation GVMultiImageKnobView
@synthesize images = mImages;

#pragma mark Bindings

- (void)bind:(NSString *)binding
    toObject:(id)observableObject
 withKeyPath:(NSString *)keyPath
     options:(NSDictionary *)options
{
    // Observe the observableObject for changes - note, pass binding identifier
    // as the context, so you get that back in observeValueForKeyPath:...
    // This way you can easily determine what needs to be updated.
    if ([binding isEqualToString:@"normalizedValue"])
    {
        [observableObject addObserver:self
                           forKeyPath:keyPath
                              options:0
                              context:NormalizedValueBindingContext];

        // Register what object and what keypath are associated with this binding
        observedObjectForNormalizedValue = observableObject;
        observedKeyPathForNormalizedValue = [keyPath copy];
    }
}

- (void)observeValueForKeyPath:(NSString *)keyPath
                      ofObject:(id)object
                        change:(NSDictionary *)change
                       context:(void *)context
{
    // You passed the binding identifier as the context when registering
    // as an observer - use that to decide what to update...
    if (context == NormalizedValueBindingContext)
    {
        id newValue = [observedObjectForNormalizedValue valueForKeyPath:observedKeyPathForNormalizedValue];
        if ((newValue == NSNoSelectionMarker) ||
            (newValue == NSNotApplicableMarker) ||
            (newValue == NSMultipleValuesMarker))
        {
            // nothing to do
        }
        else
        {
            // Handles vales class of NSNumber
            if (newValue != nil && [newValue isKindOfClass:[NSNumber class]])
                [self setNormalizedValue:newValue];
            else
                assert(false); // FIXME: add detail message
        }
    }
}

-(void)postModelUpdate
{
    if (observedObjectForNormalizedValue != nil)
    {
        [observedObjectForNormalizedValue setValue: self.normalizedValue
                                        forKeyPath: observedKeyPathForNormalizedValue];
    }
}


#pragma mark Initializations

- (void)awakeFromNib
{
    self.trackingStatus = kGVMouseTrackingStatus_Inactive;
    self.minAngle = 45;
    self.maxAngle = 315;
    self.defaultAngle = self.maxAngle;
    self.currentAngle = self.defaultAngle;
    self.lastDragPosition = NSMakePoint(0, 0);
}

- (id)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self)
    {
        mMinAngle = 0;
        mMaxAngle = 360;
        mImages = [NSMutableArray array];
    }
    return self;
}

#pragma mark Drawing

-(void) drawView:(CGContextRef)ctx
{
    assert(self.images.count > 0);
    if (self.images.count > 0)
    {
        NSImage * img = [self.images objectAtIndex:self.currentAngle/self.imageStep];
        if (img != nil)
        {
            CGImageSourceRef source = CGImageSourceCreateWithData((__bridge CFDataRef)[img TIFFRepresentation], NULL);
            CGImageRef imageRef =  CGImageSourceCreateImageAtIndex(source, 0, NULL);
            CFRelease(source);

            CGContextSaveGState (ctx);
            // Drawing using NSImage not working when Core Animation layer used.
            CGContextDrawImage (ctx, self.bounds, imageRef);
            CGContextRestoreGState (ctx);

            CGImageRelease(imageRef);
        }
    }
}

- (void)drawLayer:(CALayer *)layer inContext:(CGContextRef)ctx
{
    [self drawView:ctx];
}

- (void)drawRect:(NSRect)dirtyRect
{
    CGContextRef ctx = [NSGraphicsContext currentContext].graphicsPort;
    [self drawView:ctx];
}

#pragma mark Event Handling

- (void)mouseDown:(NSEvent *)theEvent
{
    if (self.trackingStatus == kGVMouseTrackingStatus_Inactive)
    {
        if ([theEvent modifierFlags] & NSAlternateKeyMask || theEvent.clickCount > 1)
        {
            // Resetting to default value with animation
//            [NSAnimationContext beginGrouping];
//            [[NSAnimationContext currentContext] setDuration:0.125];
//            [[self animator] setCurrentAngle: self.defaultAngle ];
//            [NSAnimationContext endGrouping];
            self.currentAngle = self.defaultAngle;
            [self postModelUpdate];
            self.trackingStatus = kGVMouseTrackingStatus_ResetToDefault;
        }
        else
        {
            // Continue tracking
            self.lastDragPosition = [self convertPoint:theEvent.locationInWindow fromView:nil];
            self.trackingStatus = kGVMouseTrackingStatus_DragActive;
        }
        [self setNeedsDisplay:YES];
    }
}

- (void)mouseDragged:(NSEvent *)event
{
    if (self.trackingStatus == kGVMouseTrackingStatus_DragActive)
    {
        NSPoint center = [self convertPoint:event.locationInWindow fromView:nil];
        CGFloat delta = (ABS(event.deltaX) > ABS(event.deltaY)) ? event.deltaX : -event.deltaY;
        if ([event modifierFlags] & NSCommandKeyMask)
            delta *= 0.1;
        
        self.lastDragPosition = center;
        self.currentAngle += delta;
        [self postModelUpdate];
        [self setNeedsDisplay:YES];
    }
}

- (void)mouseUp:(NSEvent *)theEvent
{
    if (self.trackingStatus == kGVMouseTrackingStatus_DragActive ||
        self.trackingStatus == kGVMouseTrackingStatus_ResetToDefault)
    {
        self.lastDragPosition = NSMakePoint(0, 0);
        self.trackingStatus = kGVMouseTrackingStatus_Inactive;
        [self postModelUpdate];
        [self setNeedsDisplay:YES];
    }
}

- (void)scrollWheel:(NSEvent *)theEvent
{
    if (self.trackingStatus == kGVMouseTrackingStatus_Inactive)
    {
        CGFloat delta = -theEvent.deltaY * 2; /// FIXME: Implement scroll sensitivity
        if ([theEvent modifierFlags] & NSCommandKeyMask)
            delta *= 0.1;
        
        self.currentAngle += delta;
        [self postModelUpdate];
        [self setNeedsDisplay:YES];
    }
}

#pragma mark Animation

+ (id)defaultAnimationForKey:(NSString *)key
{
    if ([key isEqualToString:@"currentAngle"])
    {
        // By default, use simple linear interpolation.
        return [CABasicAnimation animation];
    } 
    else
    {
        return [super defaultAnimationForKey:key];
    }
}

#pragma mark Setters/Getters

- (CGFloat)currentAngle
{
    return mCurrentAngle;
}

- (void)setCurrentAngle:(CGFloat)newValue
{
    CGFloat newAngle = newValue;
    if (newValue < self.minAngle)
        newAngle = self.minAngle;
    if (newValue > self.maxAngle)
        newAngle = self.maxAngle; 
    
    mCurrentAngle = newAngle;
    self.needsDisplay = YES;
}

- (CGFloat)minAngle
{
    return mMinAngle;
}

- (void)setMinAngle:(CGFloat)newValue
{
    CGFloat newAngle = newValue;
    if (newValue < 0)
        newAngle = 0;
    if (newValue > self.maxAngle)
        newAngle = self.maxAngle;
    
    mMinAngle = newAngle;
}

- (CGFloat)maxAngle
{
    return mMaxAngle;
}

- (void)setMaxAngle:(CGFloat)newValue
{
    CGFloat newAngle = newValue;
    if (newValue < self.minAngle)
        newAngle = self.minAngle;
    if (newValue > 360)
        newAngle = 360;   
    
    mMaxAngle = newAngle;
}

- (CGFloat)defaultAngle
{
    return mDefaultAngle;
}

- (void)setDefaultAngle:(CGFloat)newValue
{
    CGFloat newAngle = newValue;
    if (newValue < self.minAngle)
        newAngle = self.minAngle;
    if (newValue > self.maxAngle)
        newAngle = self.maxAngle;    
    
    mDefaultAngle = newAngle;
}

-(NSNumber*)normalizedValue
{
    CGFloat normValue = (self.currentAngle - self.minAngle)/(CGFloat)(self.maxAngle - self.minAngle);
    CGFloat exponent = log (0.1f) / log (0.5f); // Trick for logarithmic scalling
    CGFloat normValueScalled = powf (normValue, exponent);
    return [NSNumber numberWithFloat:normValueScalled];
}

-(void)setNormalizedValue:(NSNumber*)normalizedValue
{
    CGFloat exponentInv = log (0.5f) / log (0.1f); // Trick for logarithmic scalling
    CGFloat normValueScalled = powf (normalizedValue.floatValue, exponentInv);
    CGFloat plainValue = (self.maxAngle - self.minAngle) * normValueScalled + self.minAngle;
    self.currentAngle = plainValue;
    self.needsDisplay = YES;
}

@end // EOF ]]>