//
//  GVNinePartBackgroundView.m
//  GVNinePartBackgroundView
//
//  Created by Vlad Gorloff on 23.07.12.
//  Copyright (c) 2012 Vlad Gorloff. All rights reserved.
//

#import "GVNinePartBackgroundView.h"

/*
 References:
 
 - Using "install_name_tool" for application with Frameworks inside: http://doc.qt.nokia.com/qq/qq09-mac-deployment.html
 - Drawing Layer Content With Application Kit Classes: http://goo.gl/7WgOK
 - Sample code: GeekGameBoard, NineSlice
 
 */

CGImageRef GVCreateCGImageFromFile (NSString *path)
{
    NSCAssert1(path != nil, @"path=%@", path);
    CGImageRef image     = NULL;
    CFURLRef url         = (__bridge CFURLRef) [NSURL fileURLWithPath:path];
    CGImageSourceRef src = CGImageSourceCreateWithURL(url, NULL);
    if( src )
    {
        image = CGImageSourceCreateImageAtIndex(src, 0, NULL);
        CFRelease(src);
        if(!image)
            NSLog(@"Warning: CGImageSourceCreateImageAtIndex failed on file %@ (ptr size=%lu)", path, sizeof(void*));
    }
    return image;
}

// callback for CreateImagePattern.
static void _GVDrawPatternImage (void *info, CGContextRef ctx)
{
    CGImageRef image = (CGImageRef) info;
    CGContextDrawImage(ctx,
                       CGRectMake(0,0, CGImageGetWidth(image),CGImageGetHeight(image)),
                       image);
}

// callback for CreateImagePattern.
static void _GVReleasePatternImage( void *info )
{
    CGImageRelease( (CGImageRef)info );
}


CGPatternRef GVCreateImagePattern( CGImageRef image )
{
    NSCParameterAssert(image);
    long width = CGImageGetWidth(image);
    long height = CGImageGetHeight(image);
    static const CGPatternCallbacks callbacks = {0, &_GVDrawPatternImage, &_GVReleasePatternImage};
    return CGPatternCreate (image,
                            CGRectMake (0, 0, width, height),
                            CGAffineTransformMake (1, 0, 0, 1, 0, 0),
                            width,
                            height,
                            kCGPatternTilingConstantSpacing,
                            true,
                            &callbacks);
}


CGColorRef GVCreatePatternColor( CGImageRef image )
{
    CGPatternRef pattern = GVCreateImagePattern(image);
    CGColorSpaceRef space = CGColorSpaceCreatePattern(NULL);
    CGFloat components[1] = {1.0};
    CGColorRef color = CGColorCreateWithPattern(space, pattern, components);
    CGColorSpaceRelease(space);
    CGPatternRelease(pattern);
    return color;
}


typedef enum
{
    kGVDrawUsingCocoa = 1,
    kGVDrawUsingCoreGraphics
    
} KGVDrawingMethod;

@interface GVNinePartBackgroundView()
{
    // Nine part image references
    CGImageRef topLeftCorner;
    CGImageRef topEdgeFill;
    CGImageRef topRightCorner;
    
    CGImageRef leftEdgeFill;
    CGImageRef centerFill;
    CGImageRef rightEdgeFill;
    
    CGImageRef bottomLeftCorner;
    CGImageRef bottomEdgeFill;
    CGImageRef bottomRightCorner;
}
@end

@implementation GVNinePartBackgroundView

// Extracting nine images from original image, using offsets
- (void)setBackgroundImage:(NSString *)theImagePath;
{
    // Function "GVCreateCGImageFromFile" similar to "CreateCGImageFromFile" from GeekGameBoard sample code
    // Search for sample code in Xcode documentation
    CGImageRef  originalImage   = GVCreateCGImageFromFile(theImagePath);
    NSInteger   horizFillSize   = CGImageGetWidth(originalImage) - (self.leftOffset + self.rightOffset);
    NSInteger   vertFillSize    = CGImageGetHeight(originalImage) - (self.topOffset + self.bottomOffset);
    assert(horizFillSize > 0);
    assert(vertFillSize > 0);

    CGRect rect;
    rect                = CGRectMake(0, 0, self.leftOffset, self.topOffset);
    topLeftCorner       = CGImageCreateWithImageInRect(originalImage, rect);
    rect                = CGRectMake(self.leftOffset, 0, horizFillSize, self.topOffset);
    topEdgeFill         = CGImageCreateWithImageInRect(originalImage, rect);
    rect                = CGRectMake(self.leftOffset + horizFillSize, 0, self.rightOffset, self.topOffset);
    topRightCorner      = CGImageCreateWithImageInRect(originalImage, rect);
    rect                = CGRectMake(0, self.topOffset, self.leftOffset, vertFillSize);
    leftEdgeFill        = CGImageCreateWithImageInRect(originalImage, rect);
    rect                = CGRectMake(self.leftOffset, self.topOffset, horizFillSize, vertFillSize);
    centerFill          = CGImageCreateWithImageInRect(originalImage, rect);
    rect                = CGRectMake(self.leftOffset + horizFillSize, self.topOffset, self.rightOffset, vertFillSize);
    rightEdgeFill       = CGImageCreateWithImageInRect(originalImage, rect);
    rect                = CGRectMake(0, self.topOffset + vertFillSize, self.leftOffset, self.bottomOffset);
    bottomLeftCorner    = CGImageCreateWithImageInRect(originalImage, rect);
    rect                = CGRectMake(self.leftOffset, self.topOffset + vertFillSize, horizFillSize, self.bottomOffset);
    bottomEdgeFill      = CGImageCreateWithImageInRect(originalImage, rect);
    rect = CGRectMake(self.leftOffset + horizFillSize, self.topOffset + vertFillSize, self.rightOffset, self.bottomOffset);
    bottomRightCorner = CGImageCreateWithImageInRect(originalImage, rect);
    
    CFRelease(originalImage);
}

- (void)drawRect:(NSRect)dirtyRect
{
    if (topLeftCorner == nil) // Not initialized. Nothing to do.
        return;
    
    if (self.drawingMethod == kGVDrawUsingCocoa) // This provided just for demo application
    {
        NSImage *topLeftCornerNS        = [[NSImage alloc] initWithCGImage:topLeftCorner size:NSZeroSize];
        NSImage *topEdgeFillNS          = [[NSImage alloc] initWithCGImage:topEdgeFill size:NSZeroSize];
        NSImage *topRightCornerNS       = [[NSImage alloc] initWithCGImage:topRightCorner size:NSZeroSize];
        NSImage *leftEdgeFillNS         = [[NSImage alloc] initWithCGImage:leftEdgeFill size:NSZeroSize];
        NSImage *centerFillNS           = [[NSImage alloc] initWithCGImage:centerFill size:NSZeroSize];
        NSImage *rightEdgeFillNS        = [[NSImage alloc] initWithCGImage:rightEdgeFill size:NSZeroSize];
        NSImage *bottomLeftCornerNS     = [[NSImage alloc] initWithCGImage:bottomLeftCorner size:NSZeroSize];
        NSImage *bottomEdgeFillNS       = [[NSImage alloc] initWithCGImage:bottomEdgeFill size:NSZeroSize];
        NSImage *bottomRightCornerNS    = [[NSImage alloc] initWithCGImage:bottomRightCorner size:NSZeroSize];
        
        // Drawing using standard Cocoa functionality
        NSDrawNinePartImage(self.frame, topLeftCornerNS, topEdgeFillNS, topRightCornerNS,
                            leftEdgeFillNS, centerFillNS, rightEdgeFillNS,
                            bottomLeftCornerNS, bottomEdgeFillNS, bottomRightCornerNS,
                            NSCompositeSourceOver, 1.0, NO);        
    }
    else
    {
        CGContextRef ctx = [NSGraphicsContext currentContext].graphicsPort;
        CGContextSaveGState(ctx);
        CGColorRef color;
        
        NSInteger   top             = CGImageGetHeight(topEdgeFill);
        NSInteger   bottom          = CGImageGetHeight(bottomEdgeFill);
        NSInteger   left            = CGImageGetWidth(leftEdgeFill);
        NSInteger   right           = CGImageGetWidth(rightEdgeFill);
        NSInteger   horizFillSize   = self.frame.size.width - left - right;
        NSInteger   vertFillSize    = self.frame.size.height - top - bottom;
        assert(horizFillSize > 0);
        assert(vertFillSize > 0);
        
        // Drawing fills
        CGContextSetPatternPhase(ctx, CGSizeMake(left, self.frame.size.height - top));
        // Function "GVCreatePatternColor" similar to "CreatePatternColor" from GeekGameBoard sample code
        // Search for sample code in Xcode documentation
        color = GVCreatePatternColor(centerFill);
        CGContextSetFillColorWithColor(ctx, color);
        CGContextFillRect(ctx, NSMakeRect(left, bottom, horizFillSize, vertFillSize));

        CGContextSetPatternPhase(ctx, CGSizeMake(0, self.frame.size.height - top));
        color = GVCreatePatternColor(leftEdgeFill);
        CGContextSetFillColorWithColor(ctx, color);
        CGContextFillRect(ctx, NSMakeRect(0, bottom, left, vertFillSize));

        CGContextSetPatternPhase(ctx, CGSizeMake(left, self.frame.size.height));
        color = GVCreatePatternColor(topEdgeFill);
        CGContextSetFillColorWithColor(ctx, color);
        CGContextFillRect(ctx, NSMakeRect(left, bottom + vertFillSize, horizFillSize, top));

        CGContextSetPatternPhase(ctx, CGSizeMake(left + horizFillSize, self.frame.size.height - top));
        color = GVCreatePatternColor(rightEdgeFill);
        CGContextSetFillColorWithColor(ctx, color);
        CGContextFillRect(ctx, NSMakeRect(left + horizFillSize, bottom, right, vertFillSize));

        CGContextSetPatternPhase(ctx, CGSizeMake(left, bottom));
        color = GVCreatePatternColor(bottomEdgeFill);
        CGContextSetFillColorWithColor(ctx, color);
        CGContextFillRect(ctx, NSMakeRect(left, 0, horizFillSize, bottom));

        // Drawing corners
        CGContextDrawImage(ctx, NSMakeRect(0, self.frame.size.height - top, left, top), topLeftCorner);
        CGContextDrawImage(ctx, NSMakeRect(left + horizFillSize, self.frame.size.height - top, right, top), topRightCorner);
        CGContextDrawImage(ctx, NSMakeRect(left + horizFillSize, 0, right, bottom), bottomRightCorner);
        CGContextDrawImage(ctx, NSMakeRect(0, 0, left, bottom), bottomLeftCorner);

        CGContextRestoreGState(ctx); 
    }
}

// Quick way for setting all nine part offsets to similar value
- (void)setAllOffsets:(NSInteger)newValue
{
    self.leftOffset     = newValue;
    self.rightOffset    = newValue;
    self.topOffset      = newValue;
    self.bottomOffset   = newValue;
}

@end