// <![CDATA[
//  GVTextField.mm
//  GVCustomControls
//
//  Created by Vlad Gorloff on 12.08.12.
//  Copyright (c) 2012 Vlad Gorloff. All rights reserved.
//


#import "GVTextField.h"
#import <Quartz/Quartz.h>
#import "GVDecibelFormatter.h"
#import "GVCommonArithmetics.h"

static void *NormalizedValueBindingContext = (void *)@"NormalizedValue";

@interface GVTextField()
{
    GVDecibelFormatter *dbFormatter;
    id observedObjectForNormalizedValue;
    NSString *observedKeyPathForNormalizedValue;
}

-(void)privateInit;
-(void)makeUIEditable:(BOOL)isEditable;
-(void)postModelUpdate;

@property KGVMouseTrackingStatus trackingStatus;
@property NSPoint lastDragPosition;

@end

@implementation GVTextField

#pragma mark Bindings

- (void)bind:(NSString *)binding
    toObject:(id)observableObject
 withKeyPath:(NSString *)keyPath
     options:(NSDictionary *)options
{
    // Observe the observableObject for changes - note, pass binding identifier
    // as the context, so you get that back in observeValueForKeyPath:...
    // This way you can easily determine what needs to be updated.
    if ([binding isEqualToString:@"normalizedValue"])
    {
        [observableObject addObserver:self
                           forKeyPath:keyPath
                              options:0
                              context:NormalizedValueBindingContext];

        // Register what object and what keypath are associated with this binding
        observedObjectForNormalizedValue = observableObject;
        observedKeyPathForNormalizedValue = [keyPath copy];
    }
}

- (void)observeValueForKeyPath:(NSString *)keyPath
                      ofObject:(id)object
                        change:(NSDictionary *)change
                       context:(void *)context
{
    // You passed the binding identifier as the context when registering
    // as an observer - use that to decide what to update...
    if (context == NormalizedValueBindingContext)
    {
        id newValue = [observedObjectForNormalizedValue valueForKeyPath:observedKeyPathForNormalizedValue];
        if ((newValue == NSNoSelectionMarker) || 
            (newValue == NSNotApplicableMarker) ||
            (newValue == NSMultipleValuesMarker))
        {
            // nothing to do
        }
        else
        {
            // Handles input class of NSNumber
            if (newValue != nil && [newValue isKindOfClass:[NSNumber class]])
                [self setNormalizedValue:newValue];
            else
                assert(false); // FIXME: add detail message
        }
    }
}

-(void)postModelUpdate
{
    if (observedObjectForNormalizedValue != nil)
    {
        [observedObjectForNormalizedValue setValue: self.normalizedValue
                                        forKeyPath: observedKeyPathForNormalizedValue];
    }
}

#pragma mark Initialization

- (void)awakeFromNib
{
    self.trackingStatus = kGVMouseTrackingStatus_Inactive;
    self.lastDragPosition = NSMakePoint(0, 0);

    dbFormatter = [[GVDecibelFormatter alloc] init];
    [self.cell setFormatter:dbFormatter];

    self.defaultValue = GVPlaceLongValueInRange((NSInteger)0, dbFormatter.minValue, dbFormatter.maxValue);
    [self setObjectValue:[NSNumber numberWithInteger: self.defaultValue]];
}

-(void)privateInit
{
    self.normalTextColor = [NSColor whiteColor];
    self.editableTextColor = [NSColor blackColor];

    [self setAllowsEditingTextAttributes:NO];
    [self setBezeled:NO];
    [self setBordered:NO];
    [self setImportsGraphics:NO];
    [self setSelectable:NO];
    [self setRefusesFirstResponder:YES];
    self.backgroundColor = [NSColor textBackgroundColor];
}

- (id)initWithCoder:(NSCoder *)coder
{
    self = [super initWithCoder:coder];
    if (self)
    {
        [self privateInit];
        [self makeUIEditable:NO];
    }
    return self;
}

#pragma mark Event Handling

- (void)mouseDown:(NSEvent *)theEvent
{
    if(theEvent.clickCount > 1)
    {
        [self makeUIEditable:YES];
        [super mouseDown:theEvent];
    }
    else
    {
        [self.window makeFirstResponder:self];
        
        if (self.trackingStatus == kGVMouseTrackingStatus_Inactive)
        {
            if ([theEvent modifierFlags] & NSAlternateKeyMask)
            {
                // Resetting to default value with animation
//                [NSAnimationContext beginGrouping];
//                [[NSAnimationContext currentContext] setDuration:0.125];
//                [[self animator] setObjectValue:[NSNumber numberWithInteger: self.defaultValue]];
//                [NSAnimationContext endGrouping];
                self.objectValue = [NSNumber numberWithInteger: self.defaultValue];
                [self postModelUpdate];
                self.trackingStatus = kGVMouseTrackingStatus_ResetToDefault;
            }
            else
            {
                // Continue tracking
                self.lastDragPosition = [self convertPoint:theEvent.locationInWindow fromView:nil];
                self.trackingStatus = kGVMouseTrackingStatus_DragActive;
            }
        }
    }
}

- (void)mouseDragged:(NSEvent *)event
{
    if (self.trackingStatus == kGVMouseTrackingStatus_DragActive)
    {
        NSPoint center = [self convertPoint:event.locationInWindow fromView:nil];
        CGFloat delta = (ABS(event.deltaX) > ABS(event.deltaY)) ? event.deltaX : -event.deltaY;
        delta *= 0.2f; /// FIXME: Implement sensitivity
        if ([event modifierFlags] & NSCommandKeyMask)
            delta *= 0.1;

        self.lastDragPosition = center;
        CGFloat oldValue = [(NSNumber*)self.objectValue floatValue];
        oldValue = GVPlaceFloatValueInRange(oldValue + delta, (CGFloat)dbFormatter.minValue,
                   (CGFloat)dbFormatter.maxValue);
        self.objectValue = [NSNumber numberWithFloat: oldValue];
        [self postModelUpdate];
    }
}

- (void)mouseUp:(NSEvent *)theEvent
{
    if (self.trackingStatus == kGVMouseTrackingStatus_DragActive ||
        self.trackingStatus == kGVMouseTrackingStatus_ResetToDefault)
    {
        self.lastDragPosition = NSMakePoint(0, 0);
        self.trackingStatus = kGVMouseTrackingStatus_Inactive;
    }
}

- (void)scrollWheel:(NSEvent *)theEvent
{
    if (self.trackingStatus == kGVMouseTrackingStatus_Inactive)
    {
        CGFloat delta = -theEvent.deltaY * 0.2f; /// FIXME: Implement scroll sensitivity
        if ([theEvent modifierFlags] & NSCommandKeyMask)
            delta *= 0.1;

        CGFloat oldValue = [(NSNumber*)self.objectValue floatValue];
        oldValue = GVPlaceFloatValueInRange(oldValue + delta, (CGFloat)dbFormatter.minValue,
                    (CGFloat)dbFormatter.maxValue);
        self.objectValue = [NSNumber numberWithFloat: oldValue];
        [self postModelUpdate];
    }
}

#pragma mark Animation

+ (id)defaultAnimationForKey:(NSString *)key
{
    if ([key isEqualToString:@"objectValue"])
    {
        // By default, use simple linear interpolation.
        return [CABasicAnimation animation];
    }
    else
    {
        return [super defaultAnimationForKey:key];
    }
}

#pragma mark Delegates

- (void)cancelOperation:(id)sender
{
    [self abortEditing];
    [self makeUIEditable:NO];
}

- (void)textDidEndEditing:(NSNotification *)aNotification
{
    if (aNotification.name == NSTextDidEndEditingNotification)
    {
        int movementCode = [[[aNotification userInfo] objectForKey:@"NSTextMovement"] intValue];
        if(movementCode == NSReturnTextMovement || movementCode == NSTabTextMovement)
        {
            // Replacing notification cause and sending to super
            NSMutableDictionary *newUserInfo = [[NSMutableDictionary alloc] initWithDictionary:[aNotification userInfo]];
            [newUserInfo setObject:[NSNumber numberWithInt:NSIllegalTextMovement] forKey:@"NSTextMovement"];
            aNotification = [NSNotification notificationWithName:[aNotification name]
                                                          object:[aNotification object]
                                                        userInfo:newUserInfo];
            [super textDidEndEditing:aNotification];
            [self.window makeFirstResponder:self];
        }
        else
        {
            [super textDidEndEditing:aNotification];
        }
    }
    [self postModelUpdate];
    [self makeUIEditable:NO];
}

#pragma mark Helper Functions

-(void)makeUIEditable:(BOOL)isEditable
{
    [self setEditable:isEditable];
    [self setDrawsBackground:isEditable];
    self.textColor = (isEditable) ? self.editableTextColor : self.normalTextColor;
}

#pragma mark Getters/Setters

-(void)setMinValue:(NSInteger)aValue
{
    dbFormatter.minValue = aValue;
}

-(NSInteger)minValue
{
    return dbFormatter.minValue;
}

-(void)setMaxValue:(NSInteger)aValue
{
    dbFormatter.maxValue = aValue;
}

-(NSInteger)maxValue
{
    return dbFormatter.maxValue;
}

-(NSNumber*)normalizedValue
{
    CGFloat currValueInDb = [(NSNumber*)self.objectValue floatValue];
    CGFloat currValueNorm = powf(10, currValueInDb/20.0f);
    return [NSNumber numberWithFloat:currValueNorm];
}

-(void)setNormalizedValue:(NSNumber*)normalizedValue
{
    CGFloat normValue = GVPlaceFloatValueInRange(normalizedValue.floatValue, 0.000001f,
                        normalizedValue.floatValue);
    CGFloat valueInDb = 20.0f * log10f(normValue);
    self.objectValue = [NSNumber numberWithFloat: valueInDb];
}

@end // EOF ]]>
